# Imports
# Ignore Warnings 
import warnings
warnings.filterwarnings('ignore')

# Basic Imports 
import numpy as np
import pandas as pd
import time

# Preprocessing
from sklearn.model_selection import train_test_split, KFold, cross_val_score

# Metrics 
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

# ML Models
from lightgbm import LGBMRegressor 
from xgboost import XGBRegressor 
from catboost import CatBoostRegressor
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn import svm
from mlxtend.regressor import StackingCVRegressor

# Reading a CSV File
# 9015
df_NN = pd.read_csv("~/data/Melbourne_housing_pre.csv",  encoding="utf-8")

X=df_NN[['Year','YearBuilt','Distance','Lattitude','Longtitude','Propertycount',
          'Landsize','BuildingArea', 'Rooms','Bathroom', 'Car','Type_h','Type_t','Type_u']]
y=df_NN['LogPrice']
train_X, valid_X, train_y, valid_y = train_test_split(X,y, test_size = .20, random_state=42)

train_X2 = train_X.copy()
valid_X2 = valid_X.copy()

# Data standardization
mean = train_X.mean(axis=0)
train_X -= mean
std = train_X.std(axis=0)
train_X /= std
valid_X -= mean
valid_X /= std

##% evaluateRegressor
# from sklearn.metrics import mean_squared_error, mean_absolute_error
def evaluateRegressor(true,predicted,message = "Test set"):
    R2 = r2_score(true,predicted)
    print(message)
    print("R2 :", R2)
    

# 以下是调参后优化的性能，CatBoost在本数据集中的性能最优。
# CatBoost，LightGBM，GBR, XGB, Stack, Blend这些模型测试集中预测的准确率都超过了90%。

# 以下各字典用于记录各回归模型的性能与结果，以便在R语言中引用
train_acc = {}
valid_acc = {}
predict_train = {}
predict_valid = {}
timetake = {}

# SVM best---------------------------------------------------------------------------------------
t1 = time.time()
params =  {'C': 6.673350889023755, 'gamma': 0.05106238973376298}
svr_best = svm.SVR(**params) 
svr_model_best = svr_best.fit(train_X, train_y)
train_acc["SVM"] = svr_model_best.score(train_X,train_y)
valid_acc["SVM"] = svr_model_best.score(valid_X,valid_y)
predict_train["SVM"] = svr_model_best.predict(train_X)
predict_valid["SVM"] = svr_model_best.predict(valid_X)
t2 = time.time()
timetake["SVM"] = t2-t1
print("SVM: ",t2-t1, "\n")

# Random forest best ----------------------------------------------------------------------------
t1 = time.time()
params =  {'max_depth': 26, 'max_features': 4, 'max_leaf_nodes': 1781, 'max_samples': 0.9703002612349153, 'min_samples_leaf': 0, 'min_samples_split': 0, 'n_estimators': 381}
params['max_depth'] = params['max_depth']+3
params['min_samples_split'] = params['min_samples_split']+2
params['max_leaf_nodes'] = params['max_leaf_nodes']+2
params['min_samples_leaf'] = params['min_samples_leaf']+1
params['n_estimators'] = params['n_estimators']+50
params['max_features'] = params['max_features']+3
rf_best = RandomForestRegressor( random_state = 0,verbose=0, **params)
rf_model_best = rf_best.fit(train_X, train_y)
train_acc["RF"] = rf_model_best.score(train_X,train_y)
valid_acc["RF"] = rf_model_best.score(valid_X,valid_y)
predict_train["RF"] = rf_model_best.predict(train_X)
predict_valid["RF"] = rf_model_best.predict(valid_X)
t2 = time.time()
timetake["RF"] = t2-t1
print("Random forest: ",t2-t1, "\n")

# GBR best ------------------------------------------------------------------------------------
t1 = time.time()
params = {'alpha': 0.9014933457984278, 'learning_rate': 0.035668343067947715, 'max_depth': 6, 'max_features': 5, 'max_leaf_nodes': 943, 'min_impurity_decrease': 0.015183480929538314, 'min_samples_leaf': 2, 'min_samples_split': 30, 'n_estimators': 440, 'subsample': 0.6109167820106534}
params['n_estimators'] = params['n_estimators']+50
params['min_samples_split'] = params['min_samples_split']+2
params['min_samples_leaf'] = params['min_samples_leaf']+1
params['max_depth'] = params['max_depth']+3
params['max_features'] = params['max_features']+3
params['max_leaf_nodes'] = params['max_leaf_nodes']+2
gbr_best = GradientBoostingRegressor(random_state=0,verbose=0, **params)
gbr_model_best = gbr_best.fit(train_X, train_y)
train_acc["GBR"] = gbr_model_best.score(train_X,train_y)
valid_acc["GBR"] = gbr_model_best.score(valid_X,valid_y)
predict_train["GBR"] = gbr_model_best.predict(train_X)
predict_valid["GBR"] = gbr_model_best.predict(valid_X)
t2 = time.time()
timetake["GBR"] = t2-t1
print("GBR: ",t2-t1, "\n")

# XGB best --------------------------------------------------------------------------------------
t1 = time.time()
params =  {'colsample_bytree': 0.8413894273173292, 'gamma': 0.008478702584417519, 'learning_rate': 0.05508679239402551, 'max_bin': 4, 'max_depth': 5, 'min_child_weight': 24.524635200338793, 'n_estimators': 578, 'reg_alpha': 0.809791155072757, 'reg_lambda': 1.4490119256389808, 'subsample': 0.8429852720715357}
params['max_bin'] = params['max_bin']+50
params['max_depth'] = params['max_depth']+3
params['n_estimators'] = params['n_estimators']+100
xgb_best = XGBRegressor(objective ='reg:squarederror', seed = 0,verbosity=0, **params)  # CPU 4.96s/trial
xgb_model_best = xgb_best.fit(train_X, train_y)
train_acc["XGB"] = xgb_model_best.score(train_X,train_y)
valid_acc["XGB"] = xgb_model_best.score(valid_X,valid_y)
predict_train["XGB"] = xgb_model_best.predict(train_X)
predict_valid["XGB"] = xgb_model_best.predict(valid_X)
t2 = time.time()
timetake["XGB"] = t2-t1
print("XGB: ",t2-t1, "\n")

# LigthGBM best -------------------------------------------------------------------------------
t1 = time.time()
params = {'colsample_bytree': 0.5142540541056978, 'learning_rate': 0.014284678929509775, 'max_bin': 161, 'max_depth': 4, 'min_child_samples': 5, 'min_child_weight': 4.534457967283932, 'min_split_gain': 0.0006363777341674458, 'n_estimators': 2006, 'num_leaves': 93, 'reg_alpha': 0.0037820689583625278, 'reg_lambda': 2.947360470949046, 'subsample': 0.9448608935296047, 'subsample_freq': 2}
params['max_bin'] = params['max_bin']+50
params['max_depth'] = params['max_depth']+3
params['num_leaves'] = params['num_leaves']+20
params['min_child_samples'] = params['min_child_samples']+10
params['subsample_freq'] = params['subsample_freq']+1
params['n_estimators'] = params['n_estimators']+1000
lgbm_best = LGBMRegressor(seed=0, **params)
lgb_model_best = lgbm_best.fit(train_X, train_y)
train_acc["LGBM"] = lgbm_best.score(train_X,train_y)
valid_acc["LGBM"] = lgbm_best.score(valid_X,valid_y)
predict_train["LGBM"] = lgbm_best.predict(train_X)
predict_valid["LGBM"] = lgbm_best.predict(valid_X)
t2 = time.time()
timetake["LGBM"] = t2-t1
print("LightGBM: ",t2-t1, "\n")

# CatBoost best --------------------------------------------------------------------------------------
t1 = time.time()
params = {'bagging_temperature': 0.5402870554069704, 'border_count': 183, 'depth': 5, 'fold_len_multiplier': 4.43906516804156, 'iterations': 899, 'l2_leaf_reg': 8.334167765336101, 'learning_rate': 0.0997818676941431, 'random_strength': 6.564979609549752, 'rsm': 0.8975065545697877, 'subsample': 0.857395221266925}
params['border_count'] = params['border_count']+150
params['depth'] = params['depth']+2
params['iterations'] = params['iterations']+500
cat_best = CatBoostRegressor(task_type='CPU',
                             random_seed=0,
                             leaf_estimation_iterations=1,
                             max_ctr_complexity=0,                             
                             verbose=False, **params)  # CPU 44.64s/trial
cat_model_best = cat_best.fit(train_X, train_y)
train_acc["Cat"] = cat_model_best.score(train_X,train_y)
valid_acc["Cat"] = cat_model_best.score(valid_X,valid_y)
predict_train["Cat"] = cat_model_best.predict(train_X)
predict_valid["Cat"] = cat_model_best.predict(valid_X)
t2 = time.time()
timetake["Cat"] = t2-t1
print("CatBoost: ",t2-t1, "\n")

# Blend models in order to make the final predictions more robust to overfitting
t1 = time.time()

def blended_predictions(X):
    return (
            (0.20 * cat_model_best.predict(X)) + \
            (0.20 * lgb_model_best.predict(X)) + \
            (0.20 * xgb_model_best.predict(X)) + \
            (0.20 * gbr_model_best.predict(X)) + \
            (0.20 * rf_model_best.predict(X))) 

train_acc["Blend"] = r2_score(train_y, blended_predictions(train_X))
valid_acc["Blend"] = r2_score(valid_y, blended_predictions(valid_X))
predict_train["Blend"] = blended_predictions(train_X)
predict_valid["Blend"] = blended_predictions(valid_X)
t2 = time.time()
timetake["Blend"] = t2-t1
print("Blend: ",t2-t1, "\n")

def PredictTrain():
    df = pd.DataFrame({"origin":train_y, "svm":predict_train["SVM"], "rf":predict_train["RF"],\
         "gbr":predict_train["GBR"], "xgb":predict_train["XGB"],"lgbm":predict_train["LGBM"],\
         "cat":predict_train["Cat"],"blend":predict_train["Blend"]})
    return(df)


def PredictValid():
    df = pd.DataFrame({"origin":valid_y, "svm":predict_valid["SVM"], "rf":predict_valid["RF"],\
         "gbr":predict_valid["GBR"], "xgb":predict_valid["XGB"],"lgbm":predict_valid["LGBM"],\
         "cat":predict_valid["Cat"],"blend":predict_valid["Blend"]})
    return(df)


def performance():
    df = pd.DataFrame({"algo":list(train_acc.keys()),"train":list(train_acc.values()), \
         "valid":list(valid_acc.values()), "time":list(timetake.values())})
    return(df)

#df1 = PredictTrain()
#df2 = PredictValid()
#df3 = performance()
